#ifndef PSTRUCT_H
#define PSTRUCT_H
#include<eo>

// Rosetta includes
#include<core/pose/Pose.hh>
#include<core/scoring/ScoreFunction.hh>
#include<core/scoring/ScoringManager.hh>
#include<core/scoring/methods/EnergyMethodOptions.hh>
#include<core/io/pdb/pose_io.hh>
#include<core/io/pdb/pdb_dynamic_reader.hh>
#include<protocols/evaluation/ScoreEvaluator.hh>
#include<core/scoring/ScoreFunctionFactory.hh>


#include <core/sequence/util.hh>
#include <core/sequence/Sequence.hh>

#include <core/chemical/util.hh>
#include <core/chemical/ChemicalManager.hh> // To define the representation type
#include <core/chemical/ResidueTypeSet.fwd.hh> // Same


#include <core/io/pdb/file_data.hh>



using namespace std;

extern int n;
extern string pdbFile;

/*! \brief Encapsulates Rosetta's Pose structure.
 Defines some methods for stats on predicted models also.*/
class Pstruct : public EO< eoMinimizingFitness > { 

public:    
  
  typedef eoMinimizingFitness Fitness; // typedef for compatibility

  core::pose::Pose myPose; // Think if possible to make private
  core::scoring::ScoreFunctionOP scoreFunc; // Same

  /*! \brief Void constructor, used by default. 
   Unfortunately, Rosetta needs informations to create a Pose. 
  If you use this, you must call create(string, string) before running 
  prediction.*/
  Pstruct()
    {
     // create() must be called somewhere
    }
  
  Pstruct(core::pose::Pose & _pose,  core::scoring::ScoreFunction  & _scoreFunc):
    myPose(_pose),
    scoreFunc(_scoreFunc){
    angleKeys.resize(myPose.total_residue());
    fitContrib.resize(myPose.total_residue());
  }
    
    Pstruct (const Pstruct & _pstruct):EO< eoMinimizingFitness >(){ // copy constructor
      *this = _pstruct;
    }
    
    ~Pstruct(){ // destructor
    }
    
    std::ostream& operator<< (const Pstruct & _pstruct)
      {
        return std::cout << _pstruct.fitness();
      }

    void operator = (const Pstruct & _pstruct)
      {
        myPose = _pstruct.myPose;
        scoreFunc = _pstruct.scoreFunc;
        angleKeys = _pstruct.angleKeys;
        fitContrib = _pstruct.fitContrib;
        if (!_pstruct.invalid())
          {
            fitness(_pstruct.fitness());
          }
      }

    // Put any test here
    void test(){
      scoreFunc->show(cout, myPose);
    }
    
    /* void operator= (const Pstruct & _pstruct){ // copy assignment operator */
    /*     for (int i = 0; i < n ; i++){ */
    /*       solution[i] = _pstruct.solution[i]; */
    /*     } */
    /*     fitness(_pstruct.fitness()); // copy the fitness */
    /*   } */
    
    /*   eoResidue& operator[] (unsigned i){ */
    /*     return solution[i]; */
    /*   } */
    
    void setPhi(std::size_t position, core::Real value)
    {
      myPose.set_phi(position,value);
    }
    
    
    void setPsi(std::size_t position, core::Real value)
    {
      myPose.set_psi(position,value);
    }


    void setOmega(std::size_t position, core::Real value)
    {
      myPose.set_omega(position,value);
    }

    void load_sequence()
    {
      std::string sequence = core::sequence::read_fasta_file(pdbFile)[1]->sequence();
      core::chemical::make_pose_from_sequence(myPose,
                                              sequence,
                                              *( core::chemical::ChemicalManager::get_instance()->residue_type_set( core::chemical::CENTROID )));
    }

    /*! \brief Retrieve a window and a fragment given a key. */
    void retrieveFragmentFromKey(int position, int nbFrags, int & window, int & fragment)
    {
      window = angleKeys[position]/nbFrags;
      fragment = angleKeys[position]%nbFrags;
    }

    /*! \brief  Add a contribution to the score to a residue.*/
    void addToContrib(int position, double _contrib)
    {
      fitContrib[position] += _contrib;
    }

    void setAngleKey(int position, int key)
    {
      angleKeys[position] = key;
    }
    
    int getAngleKey(int position)
    {
      return angleKeys[position];
    }

    vector<int> & getAngleKeys()
      {
        return angleKeys;
      }
    
    /*! \brief  Initialize a structure from a pdb file. */
    void create_from_pdb(std::string inputFile, std::string scoreFile)
    {

      //scoreFunc->initialize_from_file("scoring/weights/score0.wts"); 
      scoreFunc = core::scoring::ScoreFunctionFactory::create_score_function(scoreFile);
      //      scoreFunc->add_weights_from_file(scoreFile); 
      core::io::pdb::centroid_pose_from_pdb(myPose,inputFile);
/*       core::io::pdb::pose_from_pdb(myPose, *(core::chemical::ChemicalManager::get_instance()->residue_type_set( core::chemical::CENTROID )) */
/* ,inputFile); */
      angleKeys.resize(myPose.total_residue());
      fitContrib.resize(myPose.total_residue());
    }

    
    void change_score(std::string scoreFile)
    {
      //scoreFunc->initialize_from_file(scoreFile);
      scoreFunc = core::scoring::ScoreFunctionFactory::create_score_function(scoreFile);

    }
    
    
    /*! \brief Initialize a structure from a fasta file. */
    void create(std::string inputFile, std::string scoreFile){
      // create and initialize a solution
      // core::scoring::ScoreFunctionOP scorefxn( NULL );
      scoreFunc = core::scoring::ScoreFunctionFactory::create_score_function(scoreFile);
      //scoreFunc->initialize_from_file(scoreFile);
      //scoreFunc.apply_patch_from_file("scoring/weights/score12.wts_patch");
      /*       scoreFunc.initialize_from_file("scoring/weights/score0.wts"); */
      std::string sequence = core::sequence::read_fasta_file(inputFile)[1]->sequence();
      core::chemical::make_pose_from_sequence(myPose,
                                              sequence,
                                              *( core::chemical::ChemicalManager::get_instance()->residue_type_set( core::chemical::CENTROID )));
      angleKeys.resize(myPose.total_residue());
      fitContrib.resize(myPose.total_residue());
    }
    
    /*! \brief  Evaluate the score.*/
    Fitness evaluate()  { // evaluate the solution
      Fitness score=0;
      score = (*scoreFunc) (myPose);
      return score;
    }

    void write_angles(const char * _file)
    {
      ofstream _stream(_file);
      _stream << fitness() << endl;
      for(core::Size i=1;i<=myPose.total_residue();i++)
        {
          _stream << myPose.phi(i) << " " << myPose.psi(i) << " " << myPose.omega(i) << endl;
        }
      _stream.close();
    }

    /*! \brief  Output keys related to torsion angles for all residues.*/
    void write_angleKeys(const char * _file)
    {
      ofstream _stream(_file);
      _stream << fitness() << endl;
      for(core::Size i=0;i<angleKeys.size();i++)
        {
          _stream << angleKeys[i] << endl;
        }
      _stream.close();
    }


    /*! \brief  Output score contribution of all residues to a file.*/
    void write_contrib(const char * _file)
    {
      ofstream _stream(_file);
      for(core::Size i=0;i<fitContrib.size();i++)
        {
          _stream << fitContrib[i] << endl;
        }
      _stream.close();
    }

    
    void write_fitness(const char * _file)
    {
      ofstream _stream(_file);
      _stream << fitness() << endl;
      _stream.close();
    }


    void write_angles_diff(char * _file)
    {
      ofstream _stream(_file);
      double tmpPhi = myPose.phi(1);
      double tmpPsi = myPose.psi(1);
      double tmpOme = myPose.omega(1);
      _stream << fitness() << endl;
      for(core::Size i=2;i<=myPose.total_residue();i++)
        {
          _stream << myPose.phi(i) - tmpPhi << " " << myPose.psi(i) - tmpPsi << " " << myPose.omega(i) - tmpOme << endl;
          tmpPhi = myPose.phi(i);
          tmpPsi = myPose.psi(i);
          tmpOme = myPose.omega(i);
        }
      _stream.close();
    }


 protected:
    vector<int> angleKeys; // no comment
    vector<double> fitContrib;
      
 };
#endif

  
