#ifndef MOVEOPERATORS_H
#define MOVEOPERATORS_H
#include "keyStruct.h"
#include "roulette_wheel.h"


class moveNext
{
 public :
  virtual ~moveNext(){};
  virtual void init()=0;
  virtual bool operator()(MoveKey & moveKey)=0;
};

class windowNext
{
 public:
  virtual ~windowNext(){};
  virtual void init()=0;
  virtual bool operator()(MoveKey & _movekey)=0;
  
};


template <class M>
class proteinMoveInit{
 public :
  void operator () (M & _mover, MoveKey & moveKey)
  {
    _mover.select_window(moveKey);
    moveKey.setFragmentRank(1);
  }
};



template <class M>
class proteinMoveNext: public moveNext{
public:
  proteinMoveNext(M & _mover):nbFrags(_mover.getNbFragments)
    {
    }
    
    void init()
    {}

    bool operator() (MoveKey & moveKey) 
    {
      core::Size _rank= moveKey.getFragmentRank();
      if (_rank < nbFrags){
        moveKey.setFragmentRank(++_rank);
        return true;
      }
      
      return false;
      
    }
    
 private :
    unsigned nbFrags;
    
};



template <class M>
class proteinRandMoveNext: public moveNext
{
 public:
  proteinRandMoveNext(M & _mover)
    {
      rsize = _mover.getNbFragments();
    }
  
  void init()
  {
    rankTable.clear();
    for(unsigned i=1; i<=rsize;i++)
       rankTable.push_back(i);
  }


  void setRSize(Size _size)
  {
    rsize = _size;
  }

  bool operator() (MoveKey & moveKey) {
    unsigned size = rankTable.size();
    if (size)
      {
        unsigned rank = (int) rng.uniform(size);
        moveKey.setFragmentRank(rankTable[rank]);
        rankTable[rank] = rankTable[size-1];
        rankTable.pop_back();
        return true;
      }
    return false;
  }

 private: 
  vector<unsigned> rankTable;
  unsigned rsize;
};




template <class M>
class proteinAllRandMoveNext: public moveNext
{
 public:
  proteinAllRandMoveNext(M & _mover)
    {
      rsize = _mover.getNbFragments();
      asize = _mover.getMaxFragments();
    }
  
  void init()
  {
    rankTable.clear();
    for(int i=1; i<=rsize;i++)
      rankTable.push_back((int) (rng.uniform(asize))+1);
  }


  void setRSize(Size _size)
  {
    rsize = _size;
  }

  bool operator() (MoveKey & moveKey) {
    unsigned size = rankTable.size();
    if (size)
      {
        unsigned rank = (int) rng.uniform(size);
        moveKey.setFragmentRank(rankTable[rank]);
        rankTable[rank] = rankTable[size-1];
        rankTable.pop_back();
        return true;
      }
    return false;
  }

 private: 
  vector<unsigned> rankTable;
  unsigned rsize;
  unsigned asize;
};




class proteinSingleEntropyWindowNext: public windowNext
{
 public:
  proteinSingleEntropyWindowNext(vector< double > & _entropy):
    entropy(_entropy)
      {
      }

    void init()
    {
    }
    
    bool operator() (MoveKey & moveKey)
    {
      int rank = roulette_wheel(entropy);
      moveKey.setWindowRank(rank+1);
      return true;
    }

 private:
  vector<double> & entropy;
};


class proteinSingleRedrawWindowNext: public windowNext
{
 public:
  proteinSingleRedrawWindowNext(unsigned _size)
    {
      residueNb = _size;
    }


  void init()
  {
  }

  

  bool operator() (MoveKey & moveKey) 
  {
    unsigned rank = (int) rng.uniform(residueNb);
    moveKey.setWindowRank(rank+1);
    return true;
  }



 private:
  unsigned residueNb;
};



class proteinSingleResidueWindowNext: public windowNext
{
 public:
  proteinSingleResidueWindowNext(unsigned _size)
    {
      residueNb = _size;
    }


  void init()
  {
    for(unsigned i=1; i<=residueNb;i++)
      rankTable.push_back(i);
  }

  

  bool operator() (MoveKey & moveKey) {
    unsigned size = rankTable.size();
    if (size)
      {
        unsigned rank = (int) rng.uniform(size);
        moveKey.setWindowRank(rankTable[rank]);
        rankTable[rank] = rankTable[size-1];
        rankTable.pop_back();
        return true;
      }
    return false;
  }



 private:
  unsigned residueNb;
  vector<unsigned> rankTable;
};



template <class M>
class proteinSingleResidueMoveNext: public moveNext
{
 public:
  proteinSingleResidueMoveNext(M & _mover, vector < vector <keyStruct> > &_probas, double _rate):
    probas(_probas),
    rate(_rate)
    {
      nbFrag = _mover.getMaxFragments();
    }
    
    void init()
    {
    }
    
    bool operator() (MoveKey & moveKey)
    {
      unsigned position = moveKey.getWindowRank();
      int rank = roulette_wheel(probas[position-1]);
      moveKey.setWindowRank(position-(int) (rank/nbFrag));
      moveKey.setFragmentRank((rank%nbFrag)+1);
      if (rng.flip(rate))
        return true;
      return false;

    }

        
 private:
    vector < vector < keyStruct > > & probas; 
    unsigned nbFrag;
    double rate;    
};


template <class M> 
class proteinSingleResidueRedrawMoveNext: public moveNext
{
 public:
  proteinSingleResidueRedrawMoveNext(M & _mover, vector < vector <keyStruct> > &_probas):
    probas(_probas)
    {
      nbFrag = _mover.getMaxFragments();
    }
  void init()
  {
  }

  bool operator() (MoveKey & moveKey)
  {
    unsigned position = moveKey.getWindowRank();
    int rank = roulette_wheel(probas[position-1]);
    moveKey.setWindowRank(position-(int) (rank/nbFrag));
    moveKey.setFragmentRank((rank%nbFrag)+1);

    return true;
  }

 private:
    vector < vector < keyStruct > > & probas; 
    unsigned nbFrag;
    double rate;    
  
};



template <class M> 
class proteinRedrawMoveNext: public moveNext
{
 public:
  proteinRedrawMoveNext(M & _mover)
    {
      msize = _mover.getNbFragments();
    }
  void init()
  {
  }

  bool operator() (MoveKey & moveKey)
  {
    moveKey.setFragmentRank((int) rng.uniform(msize)+1);
    return true;
  }

 private:
  unsigned msize;
};




template <class M> 
class proteinRedrawWindowNext: public windowNext
{
 public:
  proteinRedrawWindowNext(M & _mover)
    {
      msize = _mover.getWindowNb();
    }
  void init()
  {
  }

  bool operator() (MoveKey & moveKey)
  {
    moveKey.setWindowRank((int) rng.uniform(msize)+1);
    return true;
  }

 private:
  unsigned msize;
};


template <class M>
class proteinRandWindowNext: public windowNext
{
 public:
  proteinRandWindowNext(M & _mover, unsigned _size)
    {
      msize = _mover.getWindowNb();
      assert(_size <= msize); 
      rsize = _size;
    }
  
  void init()
  {
    unsigned rank = (int) rng.uniform(msize-rsize);
    for(unsigned i=1; i<=rsize;i++)
       rankTable.push_back(rank+i);
    }

  
  void setRSize(unsigned _size)
  {
    assert(_size <= msize); 
    rsize = _size;
  }

  bool operator() (MoveKey & moveKey) {
    unsigned size = rankTable.size();
    if (size)
      {
        unsigned rank = (int) rng.uniform(size);
        moveKey.setWindowRank(rankTable[rank]);
        rankTable[rank] = rankTable[size-1];
        rankTable.pop_back();
        return true;
      }
    return false;
  }
  
 private: 
  vector<unsigned> rankTable;
  unsigned rsize;
  unsigned msize;
};




template <class M, class P>
class proteinMoveBestImprSelect{
 public :
  typedef typename P::Fitness Fitness;

  proteinMoveBestImprSelect()
    {}

  void init (const Fitness & _fitness)
  {
    bestFitness=(Fitness)_fitness;
    
    first_time = true;
  }

  bool update(const P _solution,  const Fitness & _fit)
  {
      if (first_time ||_fit > bestFitness)
        {
          best_solution =  _solution;
          bestFitness = _fit;
          if (first_time)
            {
              first_time = false;
              return false;
            }
          
          return true;
        }
      return false;
  }
   
  bool update(const M & _moveKey, const Fitness & _fit)
    {
      if (first_time ||_fit > bestFitness)
        {
          if (_fit > bestFitness)
            {
              moveKey =  _moveKey;
              bestFitness = _fit;
            }
          if (first_time)
            {
              first_time = false;
              return false;
            }

          return true;
        }
      return false;
    }
    
  void operator() (M & _moveKey, Fitness  & _fit)
    {
      if(first_time)
        {
          std::runtime_error("[proteinMoveBestImprSelect] :no move or fitness saved, call updateMove before operateMove");
        }
      _moveKey.setMoveKey(moveKey);
      _fit = bestFitness;
    }

    
  void operator() (P & _solution, Fitness  & _fit)
    {
      if(first_time)
        {
          std::runtime_error("[TsFragmentMover] :no move or fitness saved, call updateMove before operateMove");
        }
      _solution = best_solution;
      _fit = bestFitness;
    }


  Fitness getBestFitness()
  {
    return bestFitness;
  }

 private :
  bool first_time;
  P best_solution;
  Fitness bestFitness;
  MoveKey moveKey;
};


template <class M, class P>
class singleResidueMoveBestImprSelect{
 public :
  typedef typename P::Fitness Fitness;

  singleResidueMoveBestImprSelect()
    {}

  void init (const Fitness & _fitness)
  {
    Fitness fitness;
    bestFitness=(Fitness)_fitness;
    
    first_time = true;
  }

  bool update(const M & _moveKey, const Fitness & _fit, const unsigned _rel_pos)
    {
       if (first_time ||_fit > bestFitness)
         {
           if (_fit > bestFitness)
             {
               moveKey =  _moveKey;
               bestFitness = _fit;
               rel_pos = _rel_pos;
             }
          
           if (first_time)
             {
               first_time = false;
               return false;
             }
          
           return true;
         }
       return false;
    }
    
  void operator() (M & _moveKey, Fitness  & _fit, unsigned & _rel_pos)
    {
      if(first_time)
        {
          std::runtime_error("[singleResidueMoveBestImprSelect] :no move or fitness saved, call updateMove before operateMove");
        }
      _moveKey.setMoveKey(moveKey);
      _fit = bestFitness;
      _rel_pos = rel_pos;
    }

  bool update(const M & _moveKey, const Fitness & _fit)
  {
    if (first_time ||_fit > bestFitness)
      {
        if (_fit > bestFitness)
          {
            moveKey =  _moveKey;
            bestFitness = _fit;
          }
        if (first_time)
          {
            first_time = false;
              return false;
          }
        
          return true;
      }
    return false;
  }
  
  void operator() (M & _moveKey, Fitness  & _fit)
  {
    if(first_time)
      {
        std::runtime_error("[singleResidueMoveBestImprSelect] :no move or fitness saved, call updateMove before operateMove");
      }
    _moveKey.setMoveKey(moveKey);
    _fit = bestFitness;
  }
  
  
  
  Fitness getBestFitness()
  {
    return bestFitness;
  }

 private :
  bool first_time;
  P best_solution;
  Fitness bestFitness;
  MoveKey moveKey;
  unsigned rel_pos;
};




template < class M, class S >
class proteinMoveIncrEval
{
  typedef typename M::EOType::Fitness Fitness;
 public:
  Fitness operator() (M & _mover, S & _move, Pstruct & _pstruct)
  {
    _mover(_pstruct, _move);
    return _pstruct.evaluate();
  }
};


template < class M, class S >
class singleResidueMoveIncrEval
{
  typedef typename M::EOType::Fitness Fitness;
 public:
  Fitness operator() (M & _mover, S & _move, unsigned _moved_one, Pstruct & _pstruct)
  {
    _mover(_pstruct, _move, _moved_one);
    return _pstruct.evaluate();
  }
  
  Fitness operator() (M & _mover, S & _move, Pstruct & _pstruct)
  {
    _mover(_pstruct, _move);
    return _pstruct.evaluate();
  }

};


#endif
