#ifndef PROTEINSMALLMOVER_H
#define PROTEINSMALLMOVER_H
#include<protocols/moves/BackboneMover.hh>
#include <core/util/basic.hh>
#include <core/scoring/Ramachandran.hh>
#include <numeric/NumericTraits.hh>


class proteinSmallMover: public protocols::moves::BackboneMover
{
 public :
  proteinSmallMover(): BackboneMover() 
    { 
      nmoves(10);
      Mover::type( "proteinSmallMover" ); 
    }

    utility::vector1<core::id::TorsionID_Range> torsion_id_ranges(core::pose::Pose &)
      {
	return utility::vector1<core::id::TorsionID_Range>();
      }
    
    bool make_move( core::pose::Pose & pose )
    {
      scoring::Ramachandran const & rama( scoring::ScoringManager::get_instance()->get_Ramachandran() );

      conformation::Residue const & current_rsd( pose.residue(resnum_) );
      
      old_phi_ = pose.phi(resnum_);
      new_phi_ = util::periodic_range( old_phi_ - small_angle_ + RG.uniform() * big_angle_, 360.0 );
      
      old_psi_ = pose.psi(resnum_);
      new_psi_ = util::periodic_range( old_psi_ - small_angle_ + RG.uniform() * big_angle_, 360.0 );
      
      old_rama_score_ = rama.eval_rama_score_residue( current_rsd.aa(), old_phi_, old_psi_ );
      new_rama_score_ = rama.eval_rama_score_residue( current_rsd.aa(), new_phi_, new_psi_ );
      if ( !check_rama() ) 
        {
          return( false );
        }
      
      pose.set_phi( resnum_, new_phi_ );
      pose.set_psi( resnum_, new_psi_ );
      
      return( true );
    }

    void set_angles( core::Real angle_in ) {
      big_angle_ = angle_in;
      small_angle_ = big_angle_/2.0;
      
    }

    void setup_list( core::pose::Pose & pose )
    {
      using namespace id;
      for ( int i=1, i_end = pose.total_residue(); i<= i_end; ++i ) 
        {
          conformation::Residue const & rsd( pose.residue( i ) );
          if ( rsd.is_protein() && movemap_->get( TorsionID( i, BB, phi_torsion ) ) &&
               movemap_->get( TorsionID( i, BB, psi_torsion ) ) ) 
            {
              char const ss( pose.secstruct( i ) );
              if ( angle_max_.count( ss ) ) 
                {
                  Real const mx( angle_max_.find( ss )->second );
                  if ( mx > 0.0 ) 
                    {
                      pos_list_.push_back( std::make_pair( i, mx ) );
                    }
                }
            }
        }
    }

    
    utility::vector1<core::id::DOF_ID_Range> dof_id_ranges(core::pose::Pose & pose)
      {
	Real static const pi(numeric::NumericTraits<Real>::pi());
	clear();
	setup_list( pose );
        
	utility::vector1<core::id::DOF_ID_Range> range_vector;
        
	for (core::Size i = 1; i <= pos_list_.size(); ++i) 
          {
            core::id::TorsionID phi_torsion(pos_list_[i].first, core::id::BB, core::id::phi_torsion);
            core::id::DOF_ID phi_dof(pose.conformation().dof_id_from_torsion_id(phi_torsion));
            if (phi_dof.valid()) 
              range_vector.push_back(core::id::DOF_ID_Range(phi_dof, -pi, pi));
            core::id::TorsionID psi_torsion(pos_list_[i].first, core::id::BB, core::id::psi_torsion);
            core::id::DOF_ID psi_dof(pose.conformation().dof_id_from_torsion_id(psi_torsion));
            if (psi_dof.valid()) 
              range_vector.push_back(core::id::DOF_ID_Range(psi_dof, -pi, pi));
          }

	return range_vector;
      }

    void apply( core::pose::Pose & pose)
    {
      clear();
      setup_list( pose );
      num_ = std::max( Size(1), std::min( nmoves_, pos_list_.size()/2 ) );
      if (pos_list_.size()/2)
        {
          for ( int k=1; k<= num_; ++k ) 
            {
              while ( true ) 
                {
                  ++tries_;
                  if ( tries_ > 10000 ) 
                    {
                      break;
                    }
                  std::pair< int, Real > const & random_pos( pos_list_[ static_cast< int >( RG.uniform() * pos_list_.size() + 1 ) ] );
                  resnum_ = random_pos.first;
                  set_angles( random_pos.second );
                  if ( std::find( already_moved_.begin(), already_moved_.end(), resnum_ ) !=
                       already_moved_.end() ) continue;
                  
                  if ( !make_move( pose ) ) 
                    continue;
                  already_moved_.push_back( resnum_ );
                  break;
                } 

            } 
        }
    }


    
    
};



#endif
